#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import octobot_commons.enums as enums

CONFIG_BACKTESTING = "backtesting"
CONFIG_BACKTESTING_DATA_FILES = "files"
CONFIG_ANALYSIS_ENABLED_OPTION = "post_analysis_enabled"
CONFIG_BACKTESTING_OTHER_MARKETS_STARTING_PORTFOLIO = 10000
BACKTESTING_DATA_OHLCV = "ohlcv"
BACKTESTING_DATA_TRADES = "trades"
BACKTESTING_FILE_PATH = f"{CONFIG_BACKTESTING}/data"
BACKTESTING_DATA_FILE_EXT = ".data"
BACKTESTING_DATA_FILE_TEMP_EXT = ".part"
BACKTESTING_DATA_FILE_SEPARATOR = "_"
BACKTESTING_DATA_FILE_TIME_WRITE_FORMAT = '%Y%m%d_%H%M%S'
BACKTESTING_DATA_FILE_TIME_READ_FORMAT = BACKTESTING_DATA_FILE_TIME_WRITE_FORMAT.replace("_", "")
BACKTESTING_DATA_FILE_TIME_DISPLAY_FORMAT = '%d %B %Y at %H:%M:%S'
BACKTESTING_DEFAULT_JOIN_TIMEOUT = 1800  # 30min

BACKTESTING_TIME_FRAMES_TO_DISPLAY = [enums.TimeFrames.THIRTY_MINUTES.value,
                                      enums.TimeFrames.ONE_HOUR.value,
                                      enums.TimeFrames.FOUR_HOURS.value,
                                      enums.TimeFrames.ONE_DAY.value]



#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import json
import os.path as path
import os
import aiohttp

import octobot_commons.logging as logging

import octobot_backtesting.enums as enums
import octobot_backtesting.constants as constants
import octobot_backtesting.data as data
import octobot_backtesting.importers as importers


class DataCollector:
    IMPORTER = importers.DataImporter

    def __init__(self, config,
                 path=constants.BACKTESTING_FILE_PATH,
                 data_format=enums.DataFormats.REGULAR_COLLECTOR_DATA):
        self.config = config
        self.path = path
        self.logger = logging.get_logger(self.__class__.__name__)

        self.should_stop = False
        self.file_name = data.get_backtesting_file_name(self.__class__, data_format)

        self.database = None
        self.aiohttp_session = None
        self.file_path = None
        self.temp_file_path = None
        self.finished = False
        self.in_progress = False
        self._ensure_file_path()
        self.set_file_path()

    async def initialize(self) -> None:
        pass

    async def stop(self, **kwargs) -> None:
        self.should_stop = True

    async def start(self) -> None:
        raise NotImplementedError("Start is not implemented")

    def is_in_progress(self):
        return self.in_progress

    def is_finished(self):
        return self.finished

    def _ensure_file_path(self):
        if not path.isdir(self.path):
            os.makedirs(self.path)

    def set_file_path(self) -> None:
        self.file_path = path.join(self.path, self.file_name) if self.path else self.file_name
        self.temp_file_path = self.file_path + constants.BACKTESTING_DATA_FILE_TEMP_EXT

    def create_database(self) -> None:
        if not self.database:
            self.database = data.DataBase(self.temp_file_path)

    def finalize_database(self):
        os.rename(self.temp_file_path, self.file_path)

    def create_aiohttp_session(self) -> None:
        if not self.aiohttp_session:
            self.aiohttp_session = aiohttp.ClientSession()

    async def stop_aiohttp_session(self) -> None:
        if self.aiohttp_session:
            await self.aiohttp_session.close()

    async def execute_request(self, url, params=None, headers=None):
        response = await self.aiohttp_session.get(url, params=params, headers=headers)
        if response.status != 200:
            if response.status == 502:  # bad gateway (should retry)
                self.logger.warning("Got a bad gateway error, retrying...")
                await asyncio.sleep(60)
                return await self.execute_request(url, params=params, headers=headers)
            else:
                try:
                    message = json.loads(await response.text())['message']
                except json.JSONDecodeError:
                    message = await response.text()
                self.logger.error(f"Error when requesting url {url} / "
                                  f"message : {message} / "
                                  f"code : {response.status} / "
                                  f"reason : {response.reason}")
            return None
        try:
            return json.loads(await response.text())
        except aiohttp.ClientPayloadError as e:
            self.logger.error(f"Failed to extract payload text : {e}")
            return None

    async def fetch_with_continuation(self, continuation_url_key, json_answer, headers, callback):
        if continuation_url_key in json_answer:
            answer = await self.execute_request(json_answer[continuation_url_key], headers=headers)
            if answer is None:
                return

            await callback(answer["data"])

            await self.fetch_with_continuation(continuation_url_key, answer, headers, callback)



#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_backtesting.api import data_file_converters
from octobot_backtesting.api import data_file
from octobot_backtesting.api import importer
from octobot_backtesting.api import backtesting
from octobot_backtesting.api import exchange_data_collector

from octobot_backtesting.api.data_file_converters import (
    convert_data_file,
)
from octobot_backtesting.api.data_file import (
    get_all_available_data_files,
    delete_data_file,
    get_file_description,
)
from octobot_backtesting.api.importer import (
    get_available_data_types,
    get_available_time_frames,
    get_available_symbols,
    get_data_timestamp_interval,
    stop_importer,
)
from octobot_backtesting.api.backtesting import (
    set_time_updater_interval,
    get_importers,
    get_backtesting_current_time,
    is_backtesting_enabled,
    get_backtesting_data_files,
    get_backtesting_duration,
    initialize_backtesting,
    initialize_independent_backtesting_config,
    modify_backtesting_timestamps,
    adapt_backtesting_channels,
    start_backtesting,
    stop_backtesting,
    stop_independent_backtesting,
)
from octobot_backtesting.api.exchange_data_collector import (
    exchange_historical_data_collector_factory,
    initialize_and_run_data_collector,
    stop_data_collector,
    is_data_collector_in_progress,
    get_data_collector_progress,
    is_data_collector_finished,
)

__all__ = [
    "convert_data_file",
    "get_all_available_data_files",
    "delete_data_file",
    "get_file_description",
    "get_available_data_types",
    "get_available_time_frames",
    "get_available_symbols",
    "get_data_timestamp_interval",
    "stop_importer",
    "set_time_updater_interval",
    "get_importers",
    "get_backtesting_current_time",
    "is_backtesting_enabled",
    "get_backtesting_data_files",
    "get_backtesting_duration",
    "initialize_backtesting",
    "initialize_independent_backtesting_config",
    "modify_backtesting_timestamps",
    "adapt_backtesting_channels",
    "start_backtesting",
    "stop_backtesting",
    "stop_independent_backtesting",
    "exchange_historical_data_collector_factory",
    "initialize_and_run_data_collector",
    "stop_data_collector",
    "is_data_collector_in_progress",
    "get_data_collector_progress",
    "is_data_collector_finished",
]



#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os
import typing

import octobot_backtesting.constants as constants
import octobot_backtesting.collectors as collectors
import octobot_backtesting.importers as importers
import octobot_commons.tentacles_management as tentacles_management
import octobot_commons.logging as commons_logging


async def create_importer_from_backtesting_file_name(config,
                                                     backtesting_file,
                                                     default_importer=None) -> typing.Optional[importers.DataImporter]:
    collector_klass = tentacles_management.get_deep_class_from_parent_subclasses(
        _parse_class_name_from_backtesting_file(backtesting_file), collectors.DataCollector)
    if collector_klass:
        importer_class = collector_klass.IMPORTER
    else:
        commons_logging.get_logger().debug(f"No specific exchange importer identified for '{backtesting_file}' "
                                           f"(maybe its filename has been changed). Using {default_importer.__name__}.")
        importer_class = default_importer
    importer = importer_class(config, backtesting_file) if importer_class else None

    if not importer:
        return None

    await importer.initialize()
    return importer


def get_default_importer(default_collector_class=collectors.AbstractExchangeHistoryCollector) -> importers.DataImporter:
    available_collectors = tentacles_management.get_all_classes_from_parent(default_collector_class)
    try:
        return available_collectors[0].IMPORTER
    except KeyError:
        return None


def _parse_class_name_from_backtesting_file(backtesting_file):
    return os.path.basename(backtesting_file).split(constants.BACKTESTING_DATA_FILE_SEPARATOR)[0]



#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import contextlib 
import sqlite3
import aiosqlite

import octobot_commons.logging as logging

import octobot_backtesting.enums as enums
import octobot_backtesting.errors as errors


class DataBase:
    TIMESTAMP_COLUMN = "timestamp"
    DEFAULT_ORDER_BY = TIMESTAMP_COLUMN
    DEFAULT_SORT = enums.DataBaseOrderBy.DESC.value
    DEFAULT_WHERE_OPERATION = "="
    DEFAULT_SIZE = -1
    CACHE_SIZE = 50

    def __init__(self, file_name):
        self.file_name = file_name
        self.logger = logging.get_logger(self.__class__.__name__)

        self.tables = []
        self.cache = {}

        self.connection = None

        # should never be used directly, use async with self.aio_cursor() as cursor: instead
        self._cursor_pool = []
        self._current_cursor_index = 0

    async def initialize(self):
        try:
            self.connection = await aiosqlite.connect(self.file_name)
            await self._add_cursor_in_pool()
            await self.__init_tables_list()
        except (sqlite3.OperationalError, sqlite3.DatabaseError) as e:
            raise errors.DataBaseNotExists(e)

    async def create_index(self, table, columns):
        await self.__execute_index_creation(table, '_'.join(columns), ', '.join(columns))

    async def _add_cursor_in_pool(self):
        self._cursor_pool.append(await self.connection.cursor())

    @contextlib.asynccontextmanager
    async def aio_cursor(self) -> sqlite3.Cursor:
        """
        Use this as a context manager to get a free database cursor
        :yield: A free cursor
        :return: None
        """
        if self._current_cursor_index != 0 and len(self._cursor_pool) <= self._current_cursor_index:
            await self._add_cursor_in_pool()
        self._current_cursor_index += 1
        try:
            yield self._cursor_pool[self._current_cursor_index - 1]
        finally:
            self._current_cursor_index -= 1

    async def __execute_index_creation(self, table, name, columns):
        async with self.aio_cursor() as cursor:
            await cursor.execute(f"CREATE INDEX index_{table.value}_{name} ON {table.value} ({columns})")

    async def insert(self, table, timestamp, **kwargs):
        if table.value not in self.tables:
            await self.__create_table(table, **kwargs)

        # Insert a row of data
        inserting_values = [f"'{value}'" for value in kwargs.values()]
        await self.__execute_insert(table, self.__insert_values(timestamp, ', '.join(inserting_values)))

    async def insert_all(self, table, timestamp, **kwargs):
        # TODO refactor with : cursor.executemany("INSERT INTO my_table VALUES (?,?)", values)
        if table.value not in self.tables:
            await self.__create_table(table, **kwargs)

        insert_values = []

        for index, values in enumerate(timestamp):
            # Insert a row of data
            inserting_values = \
                [f"'{value if not isinstance(value, list) else value[index]}'" for value in kwargs.values()]
            insert_values.append(self.__insert_values(values, ', '.join(inserting_values)))

        await self.__execute_insert(table, ", ".join(insert_values))

    def __insert_values(self, timestamp, inserting_values) -> str:
        return f"({timestamp}, {inserting_values})"

    async def __execute_insert(self, table, insert_items) -> None:
        async with self.aio_cursor() as cursor:
            await cursor.execute(f"INSERT INTO {table.value} VALUES {insert_items}")

        # Save (commit) the changes
        await self.connection.commit()

    async def select(self, table, size=DEFAULT_SIZE, order_by=DEFAULT_ORDER_BY, sort=DEFAULT_SORT, **kwargs):
        return await self.__execute_select(table=table,
                                           where_clauses=self.__where_clauses_from_kwargs(**kwargs),
                                           additional_clauses=self.__select_order_by(order_by, sort),
                                           size=size)

    async def select_count(self, table, selected_items=None, **kwargs):
        return await self.__execute_select(table=table,
                                           select_items=f"{self.__count(selected_items)}",
                                           where_clauses=self.__where_clauses_from_kwargs(**kwargs))

    async def select_max(self, table, max_columns, selected_items=None, group_by=None, **kwargs):
        return await self.__execute_select(table=table,
                                           select_items=f"{self.__max(max_columns)}"
                                                        f"{', ' if selected_items else ''}"
                                                        f"{self.__selected_columns(selected_items)}",
                                           where_clauses=self.__where_clauses_from_kwargs(**kwargs),
                                           group_by=self.__select_group_by(group_by) if group_by else "")

    async def select_min(self, table, min_columns, selected_items=None, group_by=None, **kwargs):
        return await self.__execute_select(table=table,
                                           select_items=f"{self.__min(min_columns)}"
                                                        f"{', ' if selected_items else ''}"
                                                        f"{self.__selected_columns(selected_items)}",
                                           where_clauses=self.__where_clauses_from_kwargs(**kwargs),
                                           group_by=self.__select_group_by(group_by) if group_by else "")

    async def select_from_timestamp(self, table, timestamps: list, operations: list,
                                    size=DEFAULT_SIZE, order_by=DEFAULT_ORDER_BY, sort=DEFAULT_SORT, use_cache=False,
                                    **kwargs):
        timestamps_where_clauses = self.__where_clauses_from_operations(keys=[self.TIMESTAMP_COLUMN] * len(timestamps),
                                                                        values=timestamps,
                                                                        operations=operations,
                                                                        should_quote_value=False)
        where_clause = self.__where_clauses_from_kwargs(**kwargs)
        final_where_close = f"{where_clause} AND " if where_clause and timestamps_where_clauses else where_clause
        final_where_close = f"{final_where_close}{timestamps_where_clauses}"
        return await self.__execute_select(table=table,
                                           where_clauses=final_where_close,
                                           additional_clauses=self.__select_order_by(order_by, sort),
                                           size=size)

    def __where_clauses_from_kwargs(self, should_quote_value=True, **kwargs) -> str:
        return self.__where_clauses_from_operations(list(kwargs.keys()), list(kwargs.values()), [],
                                                    should_quote_value=should_quote_value)

    def __where_clauses_from_operation(self, key, value, operation=DEFAULT_WHERE_OPERATION, should_quote_value=True):
        return f"{key} {operation if operation is not None else self.DEFAULT_WHERE_OPERATION} " \
               f"{self.__quote_value(value) if should_quote_value else value}"

    def __where_clauses_from_operations(self, keys, values, operations, should_quote_value=True):
        return " AND ".join([self.__where_clauses_from_operation(keys[i],
                                                                 values[i],
                                                                 operations[i] if len(operations) > i else None,
                                                                 should_quote_value=should_quote_value)
                             for i in range(len(keys))
                             if values[i] is not None])

    def __select_order_by(self, order_by, sort):
        return f"ORDER BY " \
               f"{order_by if order_by is not None else self.DEFAULT_ORDER_BY} " \
               f"{sort if sort is not None else self.DEFAULT_SORT}"

    def __select_group_by(self, group_by):
        return f"GROUP BY {group_by}"

    def __quote_value(self, value):
        return f"'{value}'"

    def __max(self, columns):
        return f"MAX({self.__selected_columns(columns)})"

    def __min(self, columns):
        return f"MIN({self.__selected_columns(columns)})"

    def __count(self, columns):
        return f"COUNT({self.__selected_columns(columns)})"

    def __selected_columns(self, columns=None):
        return ','.join(columns) if columns else ""

    async def __execute_select(self, table, select_items="*", where_clauses="", additional_clauses="", group_by="",
                               size=DEFAULT_SIZE):
        try:
            async with self.aio_cursor() as cursor:
                await cursor.execute(f"SELECT {select_items} FROM {table.value} "
                                     f"{'WHERE' if where_clauses else ''} {where_clauses} "
                                     f"{additional_clauses} {group_by}")
                return await cursor.fetchall() if size == self.DEFAULT_SIZE else await cursor.fetchmany(size)
        except sqlite3.OperationalError as e:
            if not await self.check_table_exists(table):
                raise errors.DataBaseNotExists(e)
            self.logger.error(f"An error occurred when executing select : {e}")
        return []

    async def check_table_exists(self, table) -> bool:
        async with self.aio_cursor() as cursor:
            await cursor.execute(f"SELECT name FROM sqlite_master WHERE type='table' AND name='{table.value}'")
            return await cursor.fetchall() != []

    async def check_table_not_empty(self, table) -> bool:
        async with self.aio_cursor() as cursor:
            await cursor.execute(f"SELECT count(*) FROM '{table.value}'")
            row_count = await cursor.fetchone()
            return row_count[0] != 0

    async def __create_table(self, table, with_index_on_timestamp=True, **kwargs) -> None:
        try:
            columns: list = list(kwargs.keys())
            async with self.aio_cursor() as cursor:
                await cursor.execute(
                    f"CREATE TABLE {table.value} ({self.TIMESTAMP_COLUMN} datetime, {' text, '.join([col for col in columns])})")

            if with_index_on_timestamp:
                await self.create_index(table, [self.TIMESTAMP_COLUMN])

                for i in range(1, round(len(columns) / 2) + 1):
                    await self.create_index(table, [self.TIMESTAMP_COLUMN] + [columns[u] for u in range(0, i)])

        except sqlite3.OperationalError:
            self.logger.error(f"{table} already exists")
        finally:
            self.tables.append(table.value)

    async def __init_tables_list(self):
        async with self.aio_cursor() as cursor:
            await cursor.execute(f"SELECT name FROM sqlite_master WHERE type='table'")
            self.tables = await cursor.fetchall()

    async def stop(self):
        try:
            for cursor in self._cursor_pool:
                await cursor.close()
        finally:
            if self.connection is not None:
                await self.connection.close()
                self.connection = None



#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio 

import async_channel.channels as channels
import async_channel.consumer as consumers
import async_channel.producer as producers


class TimeProducer(producers.Producer):
    def __init__(self, channel, backtesting):
        super().__init__(channel)
        self.backtesting = backtesting

    async def push(self, timestamp):
        await self.perform(timestamp)

    async def perform(self, timestamp):
        try:
            await self.backtesting.handle_time_update(timestamp)
            await self.send(timestamp)
        except asyncio.CancelledError:
            self.logger.info("Update tasks cancelled.")
        except Exception as e:
            self.logger.exception(e, True, f"Exception when triggering time update: {e}")

    async def send(self, timestamp, **kwargs):
        for consumer in self.channel.get_consumer_from_filters({}):
            await consumer.queue.put({
                "timestamp": timestamp
            })


class TimeConsumer(consumers.SupervisedConsumer):
    pass


class TimeChannel(channels.Channel):
    PRODUCER_CLASS = TimeProducer
    CONSUMER_CLASS = TimeConsumer



#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os.path as path

import octobot_backtesting.constants as constants
import octobot_backtesting.data as data
import octobot_backtesting.enums as enums


async def get_file_description(file_name, data_path=constants.BACKTESTING_FILE_PATH) -> dict:
    description = await data.get_file_description(path.join(data_path, file_name))
    if description:
        return {
            enums.DataFormatKeys.SYMBOLS.value: description[enums.DataFormatKeys.SYMBOLS.value],
            enums.DataFormatKeys.EXCHANGE.value: description[enums.DataFormatKeys.EXCHANGE.value],
            enums.DataFormatKeys.DATE.value: data.get_date(int(description[enums.DataFormatKeys.TIMESTAMP.value])),
            enums.DataFormatKeys.TIMESTAMP.value: int(description[enums.DataFormatKeys.TIMESTAMP.value]),
            enums.DataFormatKeys.START_TIMESTAMP.value: int(description[enums.DataFormatKeys.START_TIMESTAMP.value]),
            enums.DataFormatKeys.END_TIMESTAMP.value: int(description[enums.DataFormatKeys.END_TIMESTAMP.value]),
            enums.DataFormatKeys.START_DATE.value: data.get_date(
                                        int(description[enums.DataFormatKeys.START_TIMESTAMP.value])).split(" at ")[0],
            enums.DataFormatKeys.END_DATE.value: data.get_date(
                                        int(description[enums.DataFormatKeys.END_TIMESTAMP.value])).split(" at ")[0],
            enums.DataFormatKeys.TIME_FRAMES.value: [tf.value
                                                     for tf in description[enums.DataFormatKeys.TIME_FRAMES.value]],
            enums.DataFormatKeys.CANDLES_LENGTH.value: description[enums.DataFormatKeys.CANDLES_LENGTH.value],
            enums.DataFormatKeys.TYPE.value: "OctoBot data file"
        }
    else:
        return description


def get_all_available_data_files(data_path=constants.BACKTESTING_FILE_PATH) -> list:
    return data.get_all_available_data_files(data_path)


def delete_data_file(file_name, data_path=constants.BACKTESTING_FILE_PATH) -> tuple:
    return data.delete_data_file(data_path, file_name)



#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_backtesting.api as api


def _get_default_config():
    return {
        "crypto-currencies": {
            "Bitcoin": {
                "pairs": [
                    "BTC/USD"
                ]
            }
        },
        "trader-simulator": {
            "enabled": True,
            "fees": {
                "maker": 0.07,
                "taker": 0.07
            },
            "starting-portfolio": {
                "BTC": 0.5,
                "USDT": 5000
            }
        },
        "trading": {
            "reference-market": "BTC",
            "risk": 1
        }
    }


def test_is_backtesting_enabled():
    assert api.is_backtesting_enabled({}) is False
    assert api.is_backtesting_enabled({"backtesting": {}}) is False
    assert api.is_backtesting_enabled({"backtesting": {"enabled": False}}) is False
    assert api.is_backtesting_enabled({"backtesting": {"enabled": True}}) is True


def test_get_backtesting_data_files():
    assert api.get_backtesting_data_files({}) == []
    assert api.get_backtesting_data_files({"backtesting": {}}) == []
    assert api.get_backtesting_data_files({"backtesting": {"files": []}}) == []
    assert api.get_backtesting_data_files({"backtesting": {"files": ["t", "1"]}}) == ["t", "1"]



#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import json
import os.path as path
import os 
import time
from datetime import datetime

import octobot_backtesting.data as data
import octobot_commons.enums as common_enums
import octobot_commons.time_frame_manager as tmf_manager

import octobot_backtesting.constants as constants
import octobot_backtesting.enums as enums
import octobot_backtesting.errors as errors


def get_backtesting_file_name(clazz, data_format=enums.DataFormats.REGULAR_COLLECTOR_DATA):
    return f"{clazz.__name__}{constants.BACKTESTING_DATA_FILE_SEPARATOR}" \
           f"{time.time()}{get_file_ending(data_format)}"


def get_data_type(file_name):
    if file_name.endswith(constants.BACKTESTING_DATA_FILE_EXT):
        return enums.DataFormats.REGULAR_COLLECTOR_DATA


def get_file_ending(data_type):
    if data_type == enums.DataFormats.REGULAR_COLLECTOR_DATA:
        return constants.BACKTESTING_DATA_FILE_EXT


def get_date(time_info) -> str:
    """
    :param time_info: Timestamp in seconds of the time to convert
    :return: A human readable date at the backtesting data file time format
    """
    return datetime.fromtimestamp(time_info).strftime(constants.BACKTESTING_DATA_FILE_TIME_DISPLAY_FORMAT)


async def get_database_description(database):
    description = (await database.select(enums.DataTables.DESCRIPTION, size=1))[0]
    version = description[1]
    if version == "1.0":
        return {
            enums.DataFormatKeys.TIMESTAMP.value: description[0],
            enums.DataFormatKeys.VERSION.value: description[1],
            enums.DataFormatKeys.EXCHANGE.value: description[2],
            enums.DataFormatKeys.SYMBOLS.value: json.loads(description[3]),
            enums.DataFormatKeys.TIME_FRAMES.value: [common_enums.TimeFrames(tf) for tf in json.loads(description[4])],
            enums.DataFormatKeys.START_TIMESTAMP.value: 0,
            enums.DataFormatKeys.END_TIMESTAMP.value: 0,
            enums.DataFormatKeys.CANDLES_LENGTH.value:
                                    int((await database.select_count(enums.ExchangeDataTables.OHLCV, ["*"],\
                                    time_frame=tmf_manager.find_min_time_frame([common_enums.TimeFrames(tf)
                                                                    for tf in json.loads(description[4])]).value))[0][0]
                                    / len(json.loads(description[3])))
        }
    elif version == "1.1":
        return {
            enums.DataFormatKeys.TIMESTAMP.value: description[0],
            enums.DataFormatKeys.VERSION.value: description[1],
            enums.DataFormatKeys.EXCHANGE.value: description[2],
            enums.DataFormatKeys.SYMBOLS.value: json.loads(description[3]),
            enums.DataFormatKeys.TIME_FRAMES.value: [common_enums.TimeFrames(tf) for tf in json.loads(description[4])],
            enums.DataFormatKeys.START_TIMESTAMP.value: description[5],
            enums.DataFormatKeys.END_TIMESTAMP.value: description[6],
            enums.DataFormatKeys.CANDLES_LENGTH.value:
                                    int((await database.select_count(enums.ExchangeDataTables.OHLCV, ["*"],\
                                    time_frame=tmf_manager.find_min_time_frame([common_enums.TimeFrames(tf)
                                                                    for tf in json.loads(description[4])]).value))[0][0]
                                    / len(json.loads(description[3])))
        }
    else:
        raise RuntimeError(f"Unknown datafile version: {version}")


async def get_file_description(database_file):
    database = None
    try:
        database = data.DataBase(database_file)
        await database.initialize()
        description = await get_database_description(database)
    except (errors.DataBaseNotExists, TypeError):
        description = None
    finally:
        if database is not None:
            await database.stop()
    return description


def is_valid_ending(ending):
    return ending in [constants.BACKTESTING_DATA_FILE_EXT]


def get_all_available_data_files(data_collector_path):
    try:
        files = [file
                 for file in os.listdir(data_collector_path)
                 if path.isfile(path.join(data_collector_path, file)) and is_valid_ending(path.splitext(file)[1])]
    except FileNotFoundError:
        files = []
    return files


def delete_data_file(data_collector_path, file_name):
    try:
        file_path = path.join(data_collector_path, file_name)
        if path.isfile(file_path):
            os.remove(file_path)
            return True, ""
        else:
            return False, f"file can't be found"
    except Exception as e:
        return False, e



#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import abc 
import octobot_commons.logging as logging


class DataConverter:
    def __init__(self, backtesting_file_to_convert):
        self.logger = logging.get_logger(self.__class__.__name__)
        self.file_to_convert = backtesting_file_to_convert
        self.converted_file = ""

    @abc.abstractmethod
    async def can_convert(self) -> bool:
        raise NotImplementedError("can_convert is not implemented")

    @abc.abstractmethod
    async def convert(self) -> bool:
        """
        Converts self.backtesting_file_to_convert and saves the output into self.converted_file_path
        :return: True when conversion is successful, False otherwise
        """
        raise NotImplementedError("convert is not implemented")



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_backtesting.collectors as collectors


class SocialDataCollector(collectors.DataCollector):
    # IMPORTER = SocialDataImporter

    def __init__(self, config, social_name):
        super().__init__(config)
        self.social_name = social_name
        self.set_file_path()

    async def initialize(self):
        self.create_database()

        # TODO initialize with service



# cython: language_level=3
#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

cimport octobot_backtesting.data as data

cdef class DataCollector:
    cdef public dict config

    cdef public object logger
    cdef public object aiohttp_session

    cdef public bint should_stop

    cdef public str file_name
    cdef public str file_path
    cdef public str path

    cdef public bint finished
    cdef public bint in_progress

    cdef public data.DataBase database

    cpdef bint is_finished(self)
    cpdef bint is_in_progress(self)

    cpdef void create_database(self)
    cpdef void create_aiohttp_session(self)
    cpdef void set_file_path(self)

    cdef void _ensure_file_path(self)



#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_backtesting.collectors as collectors
import octobot_commons.tentacles_management as tentacles_management


def exchange_historical_data_collector_factory(exchange_name,
                                               tentacles_setup_config,
                                               symbols,
                                               time_frames=None,
                                               start_timestamp=None,
                                               end_timestamp=None) -> str:
    collector_class = tentacles_management.get_single_deepest_child_class(collectors.AbstractExchangeHistoryCollector)
    collector_instance = collector_class({}, exchange_name, tentacles_setup_config, symbols, time_frames,
                                         use_all_available_timeframes=time_frames is None,
                                         start_timestamp=start_timestamp, end_timestamp=end_timestamp)
    return collector_instance


async def initialize_and_run_data_collector(data_collector):
    await data_collector.initialize()
    await data_collector.start()
    return data_collector.file_name


async def stop_data_collector(data_collector):
    return await data_collector.stop(should_stop_database=False) if data_collector else False


def is_data_collector_in_progress(data_collector):
    return data_collector.is_in_progress() if data_collector else False


def get_data_collector_progress(data_collector):
    return (data_collector.get_current_step_index(), data_collector.get_total_steps(),
            data_collector.get_current_step_percent()) if data_collector else (0, 0, 0)


def is_data_collector_finished(data_collector):
    return not is_data_collector_in_progress(
        data_collector) and data_collector.is_finished() if data_collector else False



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest
import os
from contextlib import asynccontextmanager


from octobot_backtesting.errors import DataBaseNotExists
from octobot_backtesting.importers.exchanges.exchange_importer import ExchangeDataImporter
from octobot_backtesting.enums import ExchangeDataTables
from octobot_commons.enums import TimeFrames

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


# use context manager instead of fixture to prevent pytest threads issues
@asynccontextmanager
async def get_importer():
    database_file = os.path.join("tests", "static", "ExchangeHistoryDataCollector_1589740606.4862757.data")
    importer = ExchangeDataImporter({}, database_file)
    try:
        await importer.initialize()
        yield importer
    finally:
        await importer.stop()


async def test_initialize():
    async with get_importer() as importer:
        assert importer.exchange_name == "binance"
        assert importer.symbols == ["ETH/BTC"]
        assert importer.time_frames == [TimeFrames(tf)
                                        for tf in ("1m", "3m", "5m", "15m", "30m", "1h", "2h",
                                        "4h", "6h", "8h", "12h", "1d", "3d", "1w", "1M")]
        assert importer.available_data_types == [ExchangeDataTables.OHLCV]


async def test_get_data_timestamp_interval():
    async with get_importer() as importer:
        # over all data
        assert await importer.get_data_timestamp_interval() == (1589710680, 1590883200)
        # for 1h time frame
        assert await importer.get_data_timestamp_interval("1h") == (1587945600, 1589742000)
        # for 1M time frame
        assert await importer.get_data_timestamp_interval("1M") == (1501459200, 1590883200)


async def test_get_ohlcv():
    async with get_importer() as importer:
        # default values
        ohlcv = await importer.get_ohlcv()
        assert len(ohlcv) == 500
        # same symbol, cryptocurrency and exchange in all data (one symbol in datafile)
        exchange = ohlcv[0][1]
        cryptocurrency = ohlcv[0][2]
        symbol = ohlcv[0][3]
        assert all(data[1] == exchange for data in ohlcv)
        assert all(data[2] == cryptocurrency for data in ohlcv)
        assert all(data[3] == symbol for data in ohlcv)
        # same time frame
        assert all(data[4] == "1h" for data in ohlcv)
        # valid ohlcv data
        assert all(
            len(data[5]) == 6 and all(isinstance(element, float)
                                      for element in data[5])
            for data in ohlcv)

        # custom values
        ohlcv = await importer.get_ohlcv(exchange_name="binance", symbol="ETH/BTC",
                                         time_frame=TimeFrames.ONE_WEEK, limit=10)
        assert len(ohlcv) == 10
        assert all(data[4] == TimeFrames.ONE_WEEK.value for data in ohlcv)

        # unknown values
        ohlcv = await importer.get_ohlcv(exchange_name="binance", symbol="ETH/XXX", time_frame=TimeFrames.ONE_WEEK)
        assert len(ohlcv) == 0


async def test_get_ohlcv_from_timestamps():
    async with get_importer() as importer:
        # default values
        ohlcv = await importer.get_ohlcv_from_timestamps()
        assert len(ohlcv) == 500
        assert all(data[4] == "1h" for data in ohlcv)
        # timestamp select
        ohlcv = await importer.get_ohlcv_from_timestamps(inferior_timestamp=1587978000, superior_timestamp=1588060800)
        assert len(ohlcv) == 24
        assert all(1587978000 <= data[0] <= 1588060800 for data in ohlcv)


async def test_get_ticker():
    async with get_importer() as importer:
        # TODO complete this test when available datafile with ticker data
        with pytest.raises(DataBaseNotExists):
            await importer.get_ticker()


async def test_get_ticker_from_timestamps():
    async with get_importer() as importer:
        # TODO complete this test when available datafile with ticker data
        with pytest.raises(DataBaseNotExists):
            await importer.get_ticker_from_timestamps()


async def test_get_order_book():
    async with get_importer() as importer:
        # TODO complete this test when available datafile with order book data
        with pytest.raises(DataBaseNotExists):
            await importer.get_order_book()


async def test_get_order_book_from_timestamps():
    async with get_importer() as importer:
        # TODO complete this test when available datafile with order book data
        with pytest.raises(DataBaseNotExists):
            await importer.get_order_book_from_timestamps()


async def test_get_recent_trades():
    async with get_importer() as importer:
        # TODO complete this test when available datafile with recent trades data
        with pytest.raises(DataBaseNotExists):
            await importer.get_recent_trades()


async def test_get_recent_trades_from_timestamps():
    async with get_importer() as importer:
        # TODO complete this test when available datafile with recent trades data
        with pytest.raises(DataBaseNotExists):
            await importer.get_recent_trades_from_timestamps()


async def test_get_kline():
    async with get_importer() as importer:
        # TODO complete this test when available datafile with kline data
        with pytest.raises(DataBaseNotExists):
            await importer.get_kline()


async def test_get_kline_from_timestamps():
    async with get_importer() as importer:
        # TODO complete this test when available datafile with kline data
        with pytest.raises(DataBaseNotExists):
            await importer.get_kline_from_timestamps()



# cython: language_level=3
#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport octobot_backtesting.time.time_manager as time_manager
cimport octobot_backtesting.time.channel.time as time_channel

cdef class Backtesting:
    cdef public object config

    cdef public list backtesting_files
    cdef public list importers
    cdef public list exchange_ids

    cdef public str matrix_id

    cdef public time_manager.TimeManager time_manager
    cdef public object time_updater
    cdef public time_channel.TimeChannel time_channel
    cdef object logger

    cpdef list get_importers(self, object importer_parent_class=*)
    cpdef double get_progress(self)
    cpdef bint is_in_progress(self)

    cdef bint _has_nothing_to_do(self)



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest
import os
import asyncio
import sqlite3
import contextlib


import octobot_commons.asyncio_tools as asyncio_tools
import octobot_backtesting.errors as errors
import octobot_backtesting.data as backtesting_data
import octobot_backtesting.enums as enums

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio

DATA_FILE1 = "ExchangeHistoryDataCollector_1589740606.4862757.data"
DATA_FILE2 = "second_ExchangeHistoryDataCollector_1589740606.4862757.data"


# use context manager instead of fixture to prevent pytest threads issues
@contextlib.asynccontextmanager
async def get_database(data_file=DATA_FILE1):
    database_file = os.path.join("tests", "static", data_file)
    database = backtesting_data.DataBase(database_file)
    try:
        await database.initialize()
        yield database
    finally:
        await database.stop()


# use context manager instead of fixture to prevent pytest threads issues
@contextlib.asynccontextmanager
async def get_temp_empty_database():
    database_name = "temp_empty_database"
    database = backtesting_data.DataBase(database_name)
    try:
        await database.initialize()
        yield database
    finally:
        await database.stop()
        os.remove(database_name)


async def test_invalid_file():
    file_name = "plop"
    db = backtesting_data.DataBase(file_name)
    try:
        await db.initialize()
        assert not await db.check_table_exists(enums.ExchangeDataTables.KLINE)
        with pytest.raises(sqlite3.OperationalError):
            await db.check_table_not_empty(enums.ExchangeDataTables.KLINE)
    finally:
        await db.stop()
        os.remove(file_name)


async def test_select():
    async with get_database() as database:
        # default values
        with pytest.raises(errors.DataBaseNotExists):
            await database.select(enums.ExchangeDataTables.KLINE)

        ohlcv = await database.select(enums.ExchangeDataTables.OHLCV)
        assert len(ohlcv) == 6531

        ohlcv = await database.select(enums.ExchangeDataTables.OHLCV, time_frame="1h")
        assert len(ohlcv) == 500

        ohlcv = await database.select(enums.ExchangeDataTables.OHLCV, symbol="xyz")
        assert len(ohlcv) == 0

        ohlcv = await database.select(enums.ExchangeDataTables.OHLCV, symbol="ETH/BTC")
        assert len(ohlcv) == 6531

        changed_order_ohlcv = await database.select(enums.ExchangeDataTables.OHLCV, order_by="time_frame", symbol="ETH/BTC")
        assert changed_order_ohlcv[0] != ohlcv[0]

        ohlcv = await database.select(enums.ExchangeDataTables.OHLCV, xyz="xyz")
        assert len(ohlcv) == 0


async def test_select_max():
    async with get_database() as database:
        assert await database.select_max(enums.ExchangeDataTables.OHLCV, ["timestamp"]) == [(1590883200,)]
        assert await database.select_max(enums.ExchangeDataTables.OHLCV, ["timestamp"], time_frame="1h") == [(1589742000,)]
        assert await database.select_max(enums.ExchangeDataTables.OHLCV, ["timestamp"], ["symbol"], time_frame="1h") == \
            [(1589742000, "ETH/BTC")]


async def test_select_min():
    async with get_database() as database:
        assert await database.select_min(enums.ExchangeDataTables.OHLCV, ["timestamp"]) == [(1500249600,)]
        assert await database.select_min(enums.ExchangeDataTables.OHLCV, ["timestamp"], time_frame="1h") == [(1587945600,)]
        assert await database.select_min(enums.ExchangeDataTables.OHLCV, ["timestamp"], ["symbol"], time_frame="1h") == \
            [(1587945600, "ETH/BTC")]

async def test_select_count():
    async with get_database() as database:
        assert await database.select_count(enums.ExchangeDataTables.OHLCV, ["*"]) == [(6531,)]
        assert await database.select_count(enums.ExchangeDataTables.OHLCV, ["*"], time_frame="1h") == [(500,)]
        assert await database.select_count(enums.ExchangeDataTables.OHLCV, ["*"], time_frame="1M") == [(35,)]

async def test_select_from_timestamp():
    async with get_database() as database:
        operations = [enums.DataBaseOperations.INF_EQUALS.value]
        candles = await database.select_from_timestamp(enums.ExchangeDataTables.OHLCV, ["1587960000"], operations)
        assert len(candles) > 0
        assert all(candle[0] <= 1587960000 for candle in candles)

        operations = [enums.DataBaseOperations.INF_EQUALS.value, enums.DataBaseOperations.SUP_EQUALS.value]
        candles = await database.select_from_timestamp(enums.ExchangeDataTables.OHLCV,
                                                       ["1587960000", "1587960000"],
                                                       operations)
        assert len(candles) > 0
        assert all(candle[0] == 1587960000 for candle in candles)

        operations = [enums.DataBaseOperations.INF_EQUALS.value, enums.DataBaseOperations.SUP_EQUALS.value]
        candles = await database.select_from_timestamp(enums.ExchangeDataTables.OHLCV,
                                                       ["1587960000", "1587945600"],
                                                       operations)
        assert len(candles) == 15
        assert all(1587945600 <= candle[0] <= 1587960000 for candle in candles)

        operations = [enums.DataBaseOperations.INF_EQUALS.value, enums.DataBaseOperations.SUP_EQUALS.value]
        candles = await database.select_from_timestamp(enums.ExchangeDataTables.OHLCV,
                                                       ["1587960000", "1587945600"],
                                                       operations,
                                                       symbol="xyz")
        assert len(candles) == 0


async def test_concurrent_select():
    async with get_database() as database:
        timestamps = [ohlcv[0] for ohlcv in await database.select(enums.ExchangeDataTables.OHLCV, time_frame="1h")]
        await asyncio.gather(*[_check_select_result(database, ts) for ts in timestamps])


async def test_stop_while_concurrent_select():
    async with get_database() as database:
        timestamps = [ohlcv[0] for ohlcv in await database.select(enums.ExchangeDataTables.OHLCV, time_frame="1h")]
        await _check_select_result(database, timestamps[0])
        asyncio.create_task(asyncio.wait(asyncio.gather(*[_check_select_result(database, ts) for ts in timestamps])))
        # not enough time to finish all requests, most if not all will remaining pending
        await asyncio_tools.wait_asyncio_next_cycle()


async def test_double_database():
    async with get_database() as database1, get_database(DATA_FILE2) as database2:
        timestamps1 = [ohlcv[0] for ohlcv in await database1.select(enums.ExchangeDataTables.OHLCV, time_frame="1h")]
        timestamps2 = [ohlcv[0] for ohlcv in await database2.select(enums.ExchangeDataTables.OHLCV, time_frame="1h")]
        await asyncio.gather(*[_check_select_result(database1, ts) for ts in timestamps1])
        await asyncio.gather(*[_check_select_result(database2, ts) for ts in timestamps2])


async def test_double_database_stop_while_concurrent_select():
    async with get_database() as database1, get_database(DATA_FILE2) as database2:
        timestamps1 = [ohlcv[0] for ohlcv in await database1.select(enums.ExchangeDataTables.OHLCV, time_frame="1h")]
        timestamps2 = [ohlcv[0] for ohlcv in await database2.select(enums.ExchangeDataTables.OHLCV, time_frame="1h")]
        await _check_select_result(database1, timestamps1[0])
        await _check_select_result(database2, timestamps2[0])
        asyncio.create_task(asyncio.wait(asyncio.gather(*[_check_select_result(database1, ts) for ts in timestamps1])))
        asyncio.create_task(asyncio.wait(asyncio.gather(*[_check_select_result(database2, ts) for ts in timestamps2])))
        # not enough time to finish all requests, most if not all will remaining pending
        await asyncio_tools.wait_asyncio_next_cycle()


async def test_insert():
    async with get_temp_empty_database() as temp_empty_database:
        await temp_empty_database.insert(enums.ExchangeDataTables.OHLCV, symbol="xyz", timestamp=1, price=1, date="01")
        assert await temp_empty_database.select(enums.ExchangeDataTables.OHLCV) == [(1, 'xyz', '1', '01')]


async def test_insert_all():
    async with get_temp_empty_database() as temp_empty_database:
        await temp_empty_database.insert_all(enums.ExchangeDataTables.OHLCV,
                                             symbol=["xyz", "abc"],
                                             timestamp=[1, 2],
                                             price=[1, 10],
                                             date=["01", "05"])
        assert await temp_empty_database.select(enums.ExchangeDataTables.OHLCV) == [(2, 'abc', '10', '05'), (1, 'xyz', '1', '01')]
        assert await temp_empty_database.select(enums.ExchangeDataTables.OHLCV, date="05") == [(2, 'abc', '10', '05')]


async def test_create_index():
    async with get_temp_empty_database() as temp_empty_database:
        await temp_empty_database.insert(enums.ExchangeDataTables.OHLCV, 1, symbol="xyz", price="1", date="01")
        # ensure no exception
        await temp_empty_database.create_index(enums.ExchangeDataTables.OHLCV, ["symbol", "timestamp"])
        assert await temp_empty_database.select(enums.ExchangeDataTables.OHLCV) == [(1, 'xyz', '1', '01')]


async def _check_select_result(database, timestamp):
    ohlcv = await database.select(enums.ExchangeDataTables.OHLCV, time_frame="1h", timestamp=str(timestamp))
    assert len(ohlcv) == 1
    assert ohlcv[0][0] == timestamp



# cython: language_level=3
#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

cdef class TimeManager:
    cdef object logger

    cdef dict config
    
    cdef public double starting_timestamp
    cdef public double finishing_timestamp
    cdef public double current_timestamp
    cdef public double time_interval

    cdef public bint time_initialized

    cdef void __reset_time(self)

    cpdef void initialize(self)
    cpdef void start(self)
    cpdef bint has_finished(self)
    cpdef void next_timestamp(self)
    cpdef void set_minimum_timestamp(self, double minimum_timestamp)
    cpdef void set_maximum_timestamp(self, double maximum_timestamp)
    cpdef void set_current_timestamp(self, double timestamp)
    cpdef double get_total_iteration(self)
    cpdef double get_remaining_iteration(self)



# cython: language_level=3
#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

cimport octobot_backtesting.collectors.data_collector as data_collector


cdef class ExchangeDataCollector(data_collector.DataCollector):
    cdef public str exchange_name

    cdef public list symbols
    cdef public list time_frames
    cdef public object tentacles_setup_config
    cdef public object start_timestamp
    cdef public object end_timestamp
    cdef public int current_step_index
    cdef public int total_steps
    cdef public float current_step_percent

    cdef bint use_all_available_timeframes

    cpdef int get_current_step_index(self)
    cpdef int get_total_steps(self)
    cpdef float get_current_step_percent(self)

    cpdef void _load_timeframes_if_necessary(self)
    cpdef void _load_all_available_timeframes(self)



#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.constants as common_constants
import octobot_commons.enums as common_enums
import octobot_commons.logging as logging
import octobot_commons.time_frame_manager as time_frame_manager

import octobot_backtesting.api as api
import octobot_backtesting.errors as errors
import octobot_backtesting.backtesting as backtesting_class
import octobot_backtesting.constants as constants


async def initialize_backtesting(config, exchange_ids, matrix_id, data_files) -> backtesting_class.Backtesting:
    backtesting_instance = backtesting_class.Backtesting(config=config,
                                                         exchange_ids=exchange_ids,
                                                         matrix_id=matrix_id,
                                                         backtesting_files=data_files)
    await backtesting_instance.create_importers()
    await backtesting_instance.initialize()

    if not backtesting_instance.importers:
        raise ValueError("No importers created: did you enter the backtesting file(s) to use ?")

    return backtesting_instance


async def initialize_independent_backtesting_config(independent_backtesting) -> dict:
    return await independent_backtesting.initialize_config()


async def modify_backtesting_timestamps(backtesting, set_timestamp=None,
                                        minimum_timestamp=None, maximum_timestamp=None) -> None:
    await backtesting.time_updater.modify(set_timestamp=set_timestamp,
                                          minimum_timestamp=minimum_timestamp,
                                          maximum_timestamp=maximum_timestamp)


async def adapt_backtesting_channels(backtesting, config, importer_class, run_on_common_part_only=True,
                                     start_timestamp=None, end_timestamp=None):
    # set mininmum and maximum timestamp according to all importers data
    min_time_frame_to_consider = time_frame_manager.find_min_time_frame(
        time_frame_manager.get_config_time_frame(config))
    importers = backtesting.get_importers(importer_class)
    if not importers:
        raise RuntimeError("No exchange importer has been found for this data file, backtesting can't start.")
    try:
        timestamps = [await api.get_data_timestamp_interval(importer, min_time_frame_to_consider)
                      for importer in importers]  # [(min, max) ... ]
    except errors.MissingTimeFrame as e:
        raise RuntimeError(f"Impossible to start backtesting on this configuration: {e}")
    min_timestamps = [timestamp[0] for timestamp in timestamps]
    max_timestamps = [timestamp[1] for timestamp in timestamps]

    min_timestamp = max(min_timestamps) if run_on_common_part_only else min(min_timestamps)
    max_timestamp = min(max_timestamps) if run_on_common_part_only else max(max_timestamps)

    if min_timestamp > max_timestamp:
        raise RuntimeError(f"No candle data to run backtesting on in this time window: starting at: {min_timestamp} "
                           f"and ending at: {max_timestamp}")
    if start_timestamp is not None and end_timestamp is not None and \
            start_timestamp > end_timestamp:
        raise RuntimeError(f"No candle data to run backtesting on in this time window: starting at: {start_timestamp} "
                           f"and ending at: {end_timestamp}")

    if start_timestamp is not None:
        if min_timestamp <= start_timestamp < end_timestamp if end_timestamp else max_timestamp:
            min_timestamp = start_timestamp
        else:
            logging.get_logger("BacktestingAPI").warning(f"Can't set the minimum timestamp to {start_timestamp}. "
                                                         f"The minimum available({min_timestamp}) will be used instead.")
    if end_timestamp is not None:
        if max_timestamp >= end_timestamp > start_timestamp if start_timestamp else min_timestamp:
            max_timestamp = end_timestamp
        else:
            logging.get_logger("BacktestingAPI").warning(f"Can't set the maximum timestamp to {end_timestamp}. "
                                                         f"The maximum available({max_timestamp}) will be used instead.")

    await modify_backtesting_timestamps(
        backtesting,
        minimum_timestamp=min_timestamp,
        maximum_timestamp=max_timestamp)
    try:
        import octobot_trading.api as exchange_api

        if exchange_api.has_only_ohlcv(importers):
            set_time_updater_interval(backtesting,
                                      common_enums.TimeFramesMinutes[min_time_frame_to_consider] *
                                      common_constants.MINUTE_TO_SECONDS)
    except ImportError:
        logging.get_logger("BacktestingAPI").error("requires OctoBot-Trading package installed")


def set_time_updater_interval(backtesting, interval_in_seconds):
    backtesting.time_manager.time_interval = interval_in_seconds


async def start_backtesting(backtesting) -> None:
    await backtesting.start_time_updater()


async def stop_backtesting(backtesting) -> None:
    await backtesting.stop()


async def stop_independent_backtesting(independent_backtesting) -> None:
    await independent_backtesting.stop()


def get_importers(backtesting) -> list:
    return backtesting.importers


def get_backtesting_current_time(backtesting) -> float:
    return backtesting.time_manager.current_timestamp


def is_backtesting_enabled(config) -> bool:
    return constants.CONFIG_BACKTESTING in config \
           and common_constants.CONFIG_ENABLED_OPTION in config[constants.CONFIG_BACKTESTING] \
           and config[constants.CONFIG_BACKTESTING][common_constants.CONFIG_ENABLED_OPTION]


def get_backtesting_data_files(config) -> list:
    return config.get(constants.CONFIG_BACKTESTING, {}).get(constants.CONFIG_BACKTESTING_DATA_FILES, [])


def get_backtesting_duration(backtesting) -> float:
    return backtesting.time_updater.simulation_duration



#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import enum 


class DataFormats(enum.Enum):
    REGULAR_COLLECTOR_DATA = 0


class DataFormatKeys(enum.Enum):
    SYMBOLS = "symbols"
    EXCHANGE = "exchange"
    DATE = "date"
    TIMESTAMP = "timestamp"
    START_TIMESTAMP = "start_timestamp"
    END_TIMESTAMP = "end_timestamp"
    START_DATE = "start_date"
    END_DATE = "end_date"
    CANDLES = "candles"
    CANDLES_LENGTH = "candles_length"
    TIME_FRAMES = "time_frames"
    TYPE = "type"
    VERSION = "version"


class ReportFormat(enum.Enum):
    SYMBOL_REPORT = "symbol_report"
    BOT_REPORT = "bot_report"
    SYMBOLS_WITH_TF = "symbols_with_time_frames_frames"


class DataBaseOrderBy(enum.Enum):
    ASC = "ASC"
    DESC = "DESC"


class DataTables(enum.Enum):
    DESCRIPTION = "description"


class ExchangeDataTables(enum.Enum):
    RECENT_TRADES = "recent_trades"
    ORDER_BOOK = "order_book"
    OHLCV = "ohlcv"
    KLINE = "kline"
    TICKER = "ticker"
    FUNDING = "funding"


class DataBaseOperations(enum.Enum):
    SUP = ">"
    INF = "<"
    EQUALS = "="
    INF_EQUALS = f"{INF}{EQUALS}"
    SUP_EQUALS = f"{SUP}{EQUALS}"



#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import json

import octobot_commons.constants as common_constants
import octobot_commons.enums as common_enums

import octobot_backtesting.data as data
import octobot_backtesting.enums as enums
import octobot_backtesting.errors as errors
import octobot_backtesting.importers as importers


class ExchangeDataImporter(importers.DataImporter):
    def __init__(self, config, file_path):
        super().__init__(config, file_path)

        self.exchange_name = None
        self.symbols = []
        self.time_frames = []
        self.available_data_types = []

    async def initialize(self) -> None:
        self.load_database()
        await self.database.initialize()

        # load description
        description = await data.get_database_description(self.database)
        self.exchange_name = description[enums.DataFormatKeys.EXCHANGE.value]
        self.symbols = description[enums.DataFormatKeys.SYMBOLS.value]
        self.time_frames = description[enums.DataFormatKeys.TIME_FRAMES.value]
        await self._init_available_data_types()

        self.logger.info(f"Loaded {self.exchange_name} data file with "
                         f"{', '.join(self.symbols)} on {', '.join([tf.value for tf in self.time_frames])}")

    async def start(self) -> None:
        pass

    async def get_data_timestamp_interval(self, time_frame=None):
        minimum_timestamp: float = 0.0
        maximum_timestamp: float = 0.0

        min_ohlcv_timestamp: float = 0.0
        max_ohlcv_timestamp: float = 0.0

        for table in [enums.ExchangeDataTables.KLINE, enums.ExchangeDataTables.ORDER_BOOK,
                      enums.ExchangeDataTables.RECENT_TRADES, enums.ExchangeDataTables.TICKER]:
            if table in self.available_data_types:
                try:
                    min_timestamp = (await self.database.select_min(table, [data.DataBase.TIMESTAMP_COLUMN]))[0][0]
                    if not minimum_timestamp or minimum_timestamp > min_timestamp:
                        minimum_timestamp = min_timestamp

                    max_timestamp = (await self.database.select_max(table, [data.DataBase.TIMESTAMP_COLUMN]))[0][0]
                    if not maximum_timestamp or maximum_timestamp < max_timestamp:
                        maximum_timestamp = max_timestamp
                except (IndexError, errors.DataBaseNotExists):
                    pass

        # OHLCV timestamps
        try:
            ohlcv_kwargs = {"time_frame": time_frame} if time_frame else {}
            ohlcv_min_timestamps = (await self.database.select_min(enums.ExchangeDataTables.OHLCV,
                                                                   [data.DataBase.TIMESTAMP_COLUMN],
                                                                   [common_constants.CONFIG_TIME_FRAME],
                                                                   group_by=common_constants.CONFIG_TIME_FRAME,
                                                                   **ohlcv_kwargs
                                                                   ))

            if ohlcv_min_timestamps:
                # if the required time frame is not included in this database, ohlcv_min_timestamps is empty: ignore it
                min_ohlcv_timestamp = max(ohlcv_min_timestamps)[0]
                max_ohlcv_timestamp = (await self.database.select_max(enums.ExchangeDataTables.OHLCV,
                                                                      [data.DataBase.TIMESTAMP_COLUMN],
                                                                      **ohlcv_kwargs))[0][0]
            elif time_frame:
                raise errors.MissingTimeFrame(f"Missing time frame in data file: {time_frame}")

        except (IndexError, errors.DataBaseNotExists):
            pass

        if minimum_timestamp > 0 and maximum_timestamp > 0:
            return max(minimum_timestamp, min_ohlcv_timestamp), max(maximum_timestamp, max_ohlcv_timestamp)
        return min_ohlcv_timestamp, max_ohlcv_timestamp

    async def _init_available_data_types(self):
        self.available_data_types = [table for table in enums.ExchangeDataTables
                                     if await self.database.check_table_exists(table)
                                     and await self.database.check_table_not_empty(table)]

    def __get_operations_from_timestamps(self, superior_timestamp, inferior_timestamp):
        operations: list = []
        timestamps: list = []
        if superior_timestamp != -1:
            timestamps.append(str(superior_timestamp))
            operations.append(enums.DataBaseOperations.INF_EQUALS.value)
        if inferior_timestamp != -1:
            timestamps.append(str(inferior_timestamp))
            operations.append(enums.DataBaseOperations.SUP_EQUALS.value)

        return timestamps, operations

    def import_ohlcvs(self, ohlcvs):
        for i, val in enumerate(ohlcvs):
            ohlcvs[i] = list(val)
            # ohlcvs[i][-2] = TimeFrames(ohlcvs[i][-2])
            ohlcvs[i][-1] = json.loads(ohlcvs[i][-1])
        return ohlcvs

    async def get_ohlcv(self, exchange_name=None, symbol=None,
                        time_frame=common_enums.TimeFrames.ONE_HOUR,
                        limit=data.DataBase.DEFAULT_SIZE):
        return self.import_ohlcvs(await self.database.select(enums.ExchangeDataTables.OHLCV, size=limit,
                                                             exchange_name=exchange_name, symbol=symbol,
                                                             time_frame=time_frame.value))

    async def get_ohlcv_from_timestamps(self, exchange_name=None, symbol=None,
                                        time_frame=common_enums.TimeFrames.ONE_HOUR,
                                        limit=data.DataBase.DEFAULT_SIZE,
                                        inferior_timestamp=-1, superior_timestamp=-1) -> list:
        timestamps, operations = self.__get_operations_from_timestamps(superior_timestamp, inferior_timestamp)
        return self.import_ohlcvs(await self.database.select_from_timestamp(enums.ExchangeDataTables.OHLCV, size=limit,
                                                                            exchange_name=exchange_name, symbol=symbol,
                                                                            time_frame=time_frame.value,
                                                                            timestamps=timestamps,
                                                                            operations=operations))

    def import_tickers(self, tickers):
        for ticker in tickers:
            ticker[-1] = json.loads(ticker[-1])
        return tickers

    async def get_ticker(self, exchange_name=None, symbol=None, limit=data.DataBase.DEFAULT_SIZE):
        return self.import_tickers(await self.database.select(enums.ExchangeDataTables.TICKER, size=limit,
                                                              exchange_name=exchange_name, symbol=symbol))

    async def get_ticker_from_timestamps(self, exchange_name=None, symbol=None, limit=data.DataBase.DEFAULT_SIZE,
                                         inferior_timestamp=-1, superior_timestamp=-1):
        timestamps, operations = self.__get_operations_from_timestamps(superior_timestamp, inferior_timestamp)
        return self.import_tickers(await self.database.select_from_timestamp(enums.ExchangeDataTables.TICKER, size=limit,
                                                                             exchange_name=exchange_name, symbol=symbol,
                                                                             timestamps=timestamps,
                                                                             operations=operations))

    def import_order_books(self, order_books):
        for order_book in order_books:
            order_book[-1] = json.loads(order_book[-1])
            order_book[-2] = json.loads(order_book[-2])
        return order_books

    async def get_order_book(self, exchange_name=None, symbol=None, limit=data.DataBase.DEFAULT_SIZE):
        return self.import_order_books(await self.database.select(enums.ExchangeDataTables.ORDER_BOOK, size=limit,
                                                                  exchange_name=exchange_name, symbol=symbol))

    async def get_order_book_from_timestamps(self, exchange_name=None, symbol=None, limit=data.DataBase.DEFAULT_SIZE,
                                             inferior_timestamp=-1, superior_timestamp=-1):
        timestamps, operations = self.__get_operations_from_timestamps(superior_timestamp, inferior_timestamp)
        return self.import_order_books(
            await self.database.select_from_timestamp(enums.ExchangeDataTables.ORDER_BOOK, size=limit,
                                                      exchange_name=exchange_name, symbol=symbol,
                                                      timestamps=timestamps, operations=operations))

    def import_recent_trades(self, recent_trades):
        for recent_trade in recent_trades:
            recent_trade[-1] = json.loads(recent_trade[-1])
        return recent_trades

    async def get_recent_trades(self, exchange_name=None, symbol=None, limit=data.DataBase.DEFAULT_SIZE):
        return self.import_recent_trades(await self.database.select(enums.ExchangeDataTables.RECENT_TRADES, size=limit,
                                                                    exchange_name=exchange_name, symbol=symbol))

    async def get_recent_trades_from_timestamps(self, exchange_name=None, symbol=None, limit=data.DataBase.DEFAULT_SIZE,
                                                inferior_timestamp=-1, superior_timestamp=-1):
        timestamps, operations = self.__get_operations_from_timestamps(superior_timestamp, inferior_timestamp)
        return self.import_recent_trades(await
                                         self.database.select_from_timestamp(enums.ExchangeDataTables.RECENT_TRADES,
                                                                             size=limit,
                                                                             exchange_name=exchange_name,
                                                                             symbol=symbol,
                                                                             timestamps=timestamps,
                                                                             operations=operations))

    def import_klines(self, klines):
        for kline in klines:
            kline[-1] = json.loads(kline[-1])
        return klines

    async def get_kline(self, exchange_name=None, symbol=None,
                        time_frame=common_enums.TimeFrames.ONE_HOUR, limit=data.DataBase.DEFAULT_SIZE):
        return self.import_klines(await self.database.select(enums.ExchangeDataTables.KLINE, size=limit,
                                                             exchange_name=exchange_name, symbol=symbol,
                                                             time_frame=time_frame.value))

    async def get_kline_from_timestamps(self, exchange_name=None, symbol=None,
                                        time_frame=common_enums.TimeFrames.ONE_HOUR,
                                        limit=data.DataBase.DEFAULT_SIZE,
                                        inferior_timestamp=-1, superior_timestamp=-1):
        timestamps, operations = self.__get_operations_from_timestamps(superior_timestamp, inferior_timestamp)
        return self.import_klines(await self.database.select_from_timestamp(enums.ExchangeDataTables.KLINE, size=limit,
                                                                            exchange_name=exchange_name, symbol=symbol,
                                                                            time_frame=time_frame.value,
                                                                            timestamps=timestamps,
                                                                            operations=operations))



#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import time

import octobot_commons.logging as logging


class TimeManager:
    DEFAULT_TIMESTAMP_INIT_VALUE = -1
    DEFAULT_FINISH_TIME_DELTA = 1000
    DEFAULT_TIME_INTERVAL = 50

    def __init__(self, config):
        self.logger = logging.get_logger(self.__class__.__name__)
        self.config = config

        self.time_initialized = False

        self.starting_timestamp = self.DEFAULT_TIMESTAMP_INIT_VALUE
        self.finishing_timestamp = self.DEFAULT_TIMESTAMP_INIT_VALUE
        self.current_timestamp = self.DEFAULT_TIMESTAMP_INIT_VALUE
        self.time_interval = self.DEFAULT_TIME_INTERVAL

    def initialize(self):
        self.__reset_time()
        self.time_initialized = True

    def start(self):
        if self.starting_timestamp == self.DEFAULT_TIMESTAMP_INIT_VALUE:
            self.starting_timestamp = time.time()

        if self.finishing_timestamp == self.DEFAULT_TIMESTAMP_INIT_VALUE:
            self.finishing_timestamp = self.starting_timestamp + self.DEFAULT_FINISH_TIME_DELTA

        self.current_timestamp = self.starting_timestamp

    def __reset_time(self):
        self.set_current_timestamp(0)

    def has_finished(self):
        return self.current_timestamp >= self.finishing_timestamp

    def next_timestamp(self):
        self.current_timestamp += self.time_interval

    def set_minimum_timestamp(self, minimum_timestamp):
        if self.starting_timestamp == self.DEFAULT_TIMESTAMP_INIT_VALUE or self.starting_timestamp > minimum_timestamp:
            self.starting_timestamp = minimum_timestamp
            self.logger.info(f"Set minimum timestamp to : {minimum_timestamp}")

    def set_maximum_timestamp(self, maximum_timestamp):
        if self.finishing_timestamp == self.DEFAULT_TIMESTAMP_INIT_VALUE or \
                self.finishing_timestamp < maximum_timestamp:
            self.finishing_timestamp = maximum_timestamp
            self.logger.info(f"Set maximum timestamp to : {maximum_timestamp}")

    def set_current_timestamp(self, timestamp):
        self.current_timestamp = timestamp

    def get_total_iteration(self):
        return (self.finishing_timestamp - self.starting_timestamp) / self.time_interval

    def get_remaining_iteration(self):
        return (self.finishing_timestamp - self.current_timestamp) / self.time_interval



#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_backtesting.data import data_file_manager
from octobot_backtesting.data import database

from octobot_backtesting.data.data_file_manager import (
    get_backtesting_file_name,
    get_data_type,
    get_file_ending,
    get_date,
    is_valid_ending,
    get_all_available_data_files,
    delete_data_file,
    get_database_description,
    get_file_description,
)
from octobot_backtesting.data.database import (
    DataBase,
)

__all__ = [
    "get_backtesting_file_name",
    "get_data_type",
    "get_file_ending",
    "get_date",
    "is_valid_ending",
    "get_all_available_data_files",
    "delete_data_file",
    "get_database_description",
    "get_file_description",
    "DataBase",
]



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import json
import logging
import abc 

import time

import octobot_commons.constants as commons_constants
import octobot_backtesting.collectors.data_collector as data_collector
import octobot_backtesting.enums as enums
import octobot_backtesting.importers as importers

try:
    import octobot_trading.constants as trading_constants
except ImportError:
    logging.error("ExchangeDataCollector requires OctoBot-Trading package installed")


class ExchangeDataCollector(data_collector.DataCollector):
    VERSION = "1.1"
    IMPORTER = importers.ExchangeDataImporter

    def __init__(self, config, exchange_name, tentacles_setup_config, symbols, time_frames, use_all_available_timeframes=False,
                 data_format=enums.DataFormats.REGULAR_COLLECTOR_DATA, start_timestamp=None, end_timestamp=None):
        super().__init__(config, data_format=data_format)
        self.exchange_name = exchange_name
        self.tentacles_setup_config = tentacles_setup_config
        self.symbols = symbols if symbols else []
        self.time_frames = time_frames if time_frames else []
        self.use_all_available_timeframes = use_all_available_timeframes
        self.start_timestamp = start_timestamp
        self.end_timestamp = end_timestamp
        self.current_step_index = 0
        self.total_steps = 0
        self.current_step_percent = 0
        self.set_file_path()

    def get_current_step_index(self):
        return self.current_step_index

    def get_total_steps(self):
        return self.total_steps

    def get_current_step_percent(self):
        return self.current_step_percent

    @abc.abstractmethod
    def _load_all_available_timeframes(self):
        raise NotImplementedError("_load_all_available_timeframes is not implemented")

    async def initialize(self):
        self.create_database()
        await self.database.initialize()

        # set config from params
        self.config[commons_constants.CONFIG_TIME_FRAME] = self.time_frames
        self.config[commons_constants.CONFIG_EXCHANGES] = {self.exchange_name: {}}
        self.config[commons_constants.CONFIG_CRYPTO_CURRENCIES] = {"Symbols": {
            commons_constants.CONFIG_CRYPTO_PAIRS: self.symbols}}

    def _load_timeframes_if_necessary(self):
        if self.use_all_available_timeframes:
            self._load_all_available_timeframes()
        self.config[commons_constants.CONFIG_TIME_FRAME] = self.time_frames

    async def _create_description(self):
        await self.database.insert(enums.DataTables.DESCRIPTION,
                                   timestamp=time.time(),
                                   version=self.VERSION,
                                   exchange=self.exchange_name,
                                   symbols=json.dumps(self.symbols),
                                   time_frames=json.dumps([tf.value for tf in self.time_frames]),
                                   start_timestamp=int(self.start_timestamp/1000) if self.start_timestamp else 0,
                                   end_timestamp=int(self.end_timestamp/1000) if self.end_timestamp
                                   else int(time.time()) if self.start_timestamp else 0)

    async def save_ticker(self, timestamp, exchange, cryptocurrency, symbol, ticker, multiple=False):
        if not multiple:
            await self.database.insert(enums.ExchangeDataTables.TICKER, timestamp,
                                       exchange_name=exchange, cryptocurrency=cryptocurrency,
                                       symbol=symbol, recent_trades=json.dumps(ticker))
        else:
            await self.database.insert_all(enums.ExchangeDataTables.TICKER, timestamp,
                                           exchange_name=exchange, cryptocurrency=cryptocurrency,
                                           symbol=symbol, recent_trades=[json.dumps(t) for t in ticker])

    async def save_order_book(self, timestamp, exchange, cryptocurrency, symbol, asks, bids, multiple=False):
        if not multiple:
            await self.database.insert(enums.ExchangeDataTables.ORDER_BOOK, timestamp,
                                       exchange_name=exchange, cryptocurrency=cryptocurrency, symbol=symbol,
                                       asks=json.dumps(asks), bids=json.dumps(bids))
        else:
            await self.database.insert_all(enums.ExchangeDataTables.ORDER_BOOK, timestamp,
                                           exchange_name=exchange, cryptocurrency=cryptocurrency, symbol=symbol,
                                           asks=[json.dumps(a) for a in asks],
                                           bids=[json.dumps(b) for b in bids])

    async def save_recent_trades(self, timestamp, exchange, cryptocurrency, symbol, recent_trades, multiple=False):
        if not multiple:
            await self.database.insert(enums.ExchangeDataTables.RECENT_TRADES, timestamp,
                                       exchange_name=exchange, cryptocurrency=cryptocurrency,
                                       symbol=symbol, recent_trades=json.dumps(recent_trades))
        else:
            await self.database.insert_all(enums.ExchangeDataTables.RECENT_TRADES, timestamp,
                                           exchange_name=exchange, cryptocurrency=cryptocurrency,
                                           symbol=symbol, recent_trades=[json.dumps(rt) for rt in recent_trades])

    async def save_ohlcv(self, timestamp, exchange, cryptocurrency, symbol, time_frame, candle, multiple=False):
        if not multiple:
            await self.database.insert(enums.ExchangeDataTables.OHLCV, timestamp,
                                       exchange_name=exchange, cryptocurrency=cryptocurrency,
                                       symbol=symbol, time_frame=time_frame.value,
                                       candle=json.dumps(candle))
        else:
            await self.database.insert_all(enums.ExchangeDataTables.OHLCV, timestamp=timestamp,
                                           exchange_name=exchange, cryptocurrency=cryptocurrency,
                                           symbol=symbol, time_frame=time_frame.value,
                                           candle=[json.dumps(c) for c in candle])

    async def save_kline(self, timestamp, exchange, cryptocurrency, symbol, time_frame, kline, multiple=False):
        if not multiple:
            await self.database.insert(enums.ExchangeDataTables.KLINE, timestamp,
                                       exchange_name=exchange, cryptocurrency=cryptocurrency,
                                       symbol=symbol, time_frame=time_frame.value,
                                       candle=json.dumps(kline))
        else:
            await self.database.insert_all(enums.ExchangeDataTables.KLINE, timestamp=timestamp,
                                           exchange_name=exchange, cryptocurrency=cryptocurrency,
                                           symbol=symbol, time_frame=time_frame.value,
                                           candle=[json.dumps(kl) for kl in kline])



#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os.path as path
import typing

import octobot_backtesting.converters as converters
import octobot_commons.tentacles_management as tentacles_management


async def convert_data_file(data_file_path) -> typing.Optional[str]:
    if data_file_path and path.isfile(data_file_path):
        converter_classes = tentacles_management.get_all_classes_from_parent(converters.DataConverter)
        for converter_class in converter_classes:
            converter = converter_class(data_file_path)
            if await converter.can_convert():
                if await converter.convert():
                    return converter.converted_file
    return None



#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio

import async_channel.channels as channels
import async_channel.enums as channel_enums

import octobot_commons.channels_name as channels_name
import octobot_commons.list_util as list_util
import octobot_commons.logging as logging


class ChannelsManager:
    DEFAULT_REFRESH_TIMEOUT = 5

    def __init__(self, exchange_ids, matrix_id, refresh_timeout=DEFAULT_REFRESH_TIMEOUT):
        self.logger = logging.get_logger(self.__class__.__name__)
        self.exchange_ids = exchange_ids
        self.matrix_id = matrix_id
        self.refresh_timeout = refresh_timeout
        self.producers = []

    async def initialize(self) -> None:
        """
        Initialize Backtesting channels manager
        """
        self.logger.debug("Initializing producers...")
        try:
            self.producers = list_util.flatten_list(_get_backtesting_producers() +
                                                    self._get_trading_producers() +
                                                    self._get_evaluator_producers())

            # Initialize all producers by calling producer.start()
            for producer in list_util.flatten_list(self._get_trading_producers() + self._get_evaluator_producers()):
                await producer.start()
        except Exception as exception:
            self.logger.exception(exception, True, f"Error when initializing backtesting: {exception}")
            raise

    async def handle_new_iteration(self) -> None:
        for level_key in channel_enums.ChannelConsumerPriorityLevels:
            try:
                await asyncio.wait_for(self.refresh_priority_level(level_key.value, True), timeout=self.refresh_timeout)
            except asyncio.TimeoutError:
                self.logger.error(f"Refreshing priority level {level_key.value} has been timed out.")

    async def refresh_priority_level(self, priority_level: int, join_consumers: bool) -> None:
        while not _check_producers_consumers_emptiness(self.producers, priority_level):
            for producer in self.producers:
                await producer.synchronized_perform_consumers_queue(priority_level, join_consumers,
                                                                    self.refresh_timeout)

    def flush(self):
        self.producers = []

    def _get_trading_producers(self):
        import octobot_trading.exchange_channel as exchange_channel
        return [
            _get_channel_producers(exchange_channel.get_chan(channel_name.value, exchange_id))
            for exchange_id in self.exchange_ids
            for channel_name in channels_name.OctoBotTradingChannelsName
        ]

    def _get_evaluator_producers(self):
        import octobot_evaluators.evaluators.channel as evaluators_channel
        return [
            _get_channel_producers(evaluators_channel.get_chan(channel_name.value, self.matrix_id))
            for channel_name in channels_name.OctoBotEvaluatorsChannelsName
        ]


def _get_channel_producers(channel):
    if channel.producers:
        return channel.producers
    return [channel.get_internal_producer()]


def _get_backtesting_producers():
    return [
        _get_channel_producers(channels.get_chan(channel_name.value))
        for channel_name in channels_name.OctoBotBacktestingChannelsName
    ]


def _check_producers_consumers_emptiness(producers, priority_level):
    for producer in producers:
        if not producer.is_consumers_queue_empty(priority_level):
            return False
    return True



#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os.path as path

import octobot_commons.logging as logging

import octobot_backtesting.constants as constants
import octobot_backtesting.data as data
import octobot_backtesting.errors as errors


class DataImporter:
    def __init__(self, config, file_path):
        self.config = config
        self.file_path = file_path
        self.logger = logging.get_logger(self.__class__.__name__)

        self.should_stop = False

        self.version = None
        self.database = None

    async def initialize(self) -> None:
        pass

    async def get_data_timestamp_interval(self, time_frame=None):
        raise NotImplementedError("get_data_timestamp_interval is not implemented")

    async def stop(self) -> None:
        if not self.should_stop:
            self.should_stop = True
            await self.database.stop()

    async def start(self) -> None:
        raise NotImplementedError("Start is not implemented")

    def load_database(self) -> None:
        file_path = self.adapt_file_path_if_necessary()
        if not self.database:
            self.database = data.DataBase(file_path)

    def adapt_file_path_if_necessary(self):
        if path.isfile(self.file_path):
            return self.file_path
        else:
            candidate_path = path.join(constants.BACKTESTING_FILE_PATH, self.file_path)
            if path.isfile(candidate_path):
                return candidate_path
        raise errors.BacktestingFileNotFound(f"File {self.file_path} not found")



#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import time

import octobot_backtesting.channels_manager as channels_manager
import octobot_backtesting.time.channel.time as time_channel


class TimeUpdater(time_channel.TimeProducer):
    def __init__(self, channel, backtesting):
        super().__init__(channel, backtesting)
        self.backtesting = backtesting
        self.time_manager = backtesting.time_manager
        self.starting_time = time.time()
        self.simulation_duration = 0
        self.finished_event = asyncio.Event()

        self.channels_manager = None

    async def start(self):
        self.channels_manager = channels_manager.ChannelsManager(exchange_ids=self.backtesting.exchange_ids,
                                                                 matrix_id=self.backtesting.matrix_id)
        await self.channels_manager.initialize()
        while not self.should_stop:
            try:
                current_timestamp = self.time_manager.current_timestamp
                await self.push(self.time_manager.current_timestamp)

                self.logger.info(f"Progress : {round(min(self.backtesting.get_progress(), 1) * 100, 2)}% "
                                 f"[{current_timestamp}]")

                # Call synchronous channels callbacks
                await self.channels_manager.handle_new_iteration()

                if self.time_manager.has_finished():
                    self.logger.debug("Maximum timestamp hit, stopping...")
                    self.simulation_duration = time.time() - self.starting_time
                    self.logger.info(f"Lasted {round(self.simulation_duration, 3)}s")
                    await self.stop()
                else:
                    # jump to the next time point
                    self.time_manager.next_timestamp()
            except Exception as e:
                self.logger.exception(e, True, f"Fail to update time : {e}")
        await self.backtesting.delete_time_channel()
        self.channels_manager.flush()
        self.finished_event.set()

    async def modify(self, set_timestamp=None, minimum_timestamp=None, maximum_timestamp=None) -> None:
        if set_timestamp is not None:
            self.time_manager.set_current_timestamp(set_timestamp)

        if minimum_timestamp is not None:
            self.time_manager.set_minimum_timestamp(minimum_timestamp)
            self.time_manager.set_current_timestamp(minimum_timestamp)

        if maximum_timestamp is not None:
            self.time_manager.set_maximum_timestamp(maximum_timestamp)

    async def run(self) -> None:
        """
        Overrides default producer run() because producer task wont be created in synchronized context
        """
        await self.channel.register_producer(self)
        self.create_task()



#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.


def get_available_data_types(importer) -> list:
    return importer.available_data_types


def get_available_time_frames(exchange_importer) -> list:
    return exchange_importer.time_frames


def get_available_symbols(exchange_importer) -> list:
    return exchange_importer.symbols


async def get_data_timestamp_interval(exchange_importer, time_frame=None) -> (float, float):
    time_frame_value = time_frame.value if time_frame is not None else None
    return await exchange_importer.get_data_timestamp_interval(time_frame=time_frame_value)


async def stop_importer(importer) -> None:
    await importer.stop()



# cython: language_level=3
#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_backtesting.collectors cimport data_collector
from octobot_backtesting.collectors.data_collector cimport (
    DataCollector,
)

from octobot_backtesting.collectors cimport exchanges
from octobot_backtesting.collectors.exchanges cimport (
    ExchangeDataCollector,
    AbstractExchangeHistoryCollector,
    AbstractExchangeLiveCollector,
)

from octobot_backtesting.collectors cimport social
from octobot_backtesting.collectors.social cimport (
    SocialDataCollector,
)


__all__ = [
    "DataCollector",
    "ExchangeDataCollector",
    "AbstractExchangeHistoryCollector",
    "AbstractExchangeLiveCollector",
    "SocialDataCollector",
]



#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import async_channel.channels as channels
import async_channel.util as channel_util

import octobot_commons.logging as logging
import octobot_commons.channels_name as channels_name
import octobot_commons.tentacles_management as tentacles_management

import octobot_backtesting.util as backtesting_util
import octobot_backtesting.time as backtesting_time


class Backtesting:
    def __init__(self, config, exchange_ids, matrix_id, backtesting_files):
        self.config = config
        self.backtesting_files = backtesting_files
        self.logger = logging.get_logger(self.__class__.__name__)

        self.exchange_ids = exchange_ids
        self.matrix_id = matrix_id

        self.importers = []
        self.time_manager = None
        self.time_updater = None
        self.time_channel = None

    async def initialize(self):
        try:
            self.time_manager = backtesting_time.TimeManager(config=self.config)
            self.time_manager.initialize()

            self.time_channel = await channel_util.create_channel_instance(backtesting_time.TimeChannel,
                                                                           channels.set_chan,
                                                                           is_synchronized=True)

            self.time_updater = backtesting_time.TimeUpdater(
                channels.get_chan(channels_name.OctoBotBacktestingChannelsName.TIME_CHANNEL.value), self)
        except Exception as e:
            self.logger.exception(e, True, f"Error when initializing backtesting : {e}.")

    async def stop(self):
        await self.delete_time_channel()
        self.time_updater.backtesting = None

    async def delete_time_channel(self):
        await self.time_channel.stop()
        for consumer in self.time_channel.consumers:
            await self.time_channel.remove_consumer(consumer)
        self.time_channel.flush()
        channels.del_chan(self.time_channel.get_name())

    async def start_time_updater(self):
        await self.time_updater.run()

    async def create_importers(self):
        try:
            self.importers = [await backtesting_util.create_importer_from_backtesting_file_name(
                self.config,
                backtesting_file,
                backtesting_util.get_default_importer()
            )
                              for backtesting_file in self.backtesting_files]
        except TypeError:
            pass

    async def handle_time_update(self, timestamp):
        if self.time_manager:
            self.time_manager.set_current_timestamp(timestamp)

    def get_importers(self, importer_parent_class=None) -> list:
        return [importer
                for importer in self.importers
                if tentacles_management.default_parents_inspection(importer.__class__, importer_parent_class)] \
            if importer_parent_class is not None else self.importers

    def get_progress(self):
        if self._has_nothing_to_do():
            return 0
        return 1 - (self.time_manager.get_remaining_iteration() / self.time_manager.get_total_iteration())

    def is_in_progress(self):
        if self._has_nothing_to_do():
            return False
        else:
            return self.time_manager.get_remaining_iteration() > 0

    def _has_nothing_to_do(self):
        return not self.time_manager or self.time_manager.get_total_iteration() == 0
